/***********************************************************
 * $Id: LocalizedComparatorFactory.java 87 2007-12-17 01:06:40Z lech $
 * 
 * Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: Feb 28, 2007
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.comparators;

import java.text.Collator;
import java.util.Comparator;
import java.util.Locale;

/**
 * This ComparatorFactory implementation returns a localized {@link Collator} 
 * instance for String objects and an ComparableComparator instance for
 * Comparables.
 * 
 * @author wglas
 */
public class LocalizedComparatorFactory extends DefaultComparatorFactory {

    private Collator collator;
    
    /**
     * Construct a LocalizedComparatorFactory for the current locale and
     * the default strength {@link Collator#IDENTICAL}.
     * 
     * @see Collator#getInstance()
     */
    public LocalizedComparatorFactory()
    {
        this.collator = Collator.getInstance();
        this.collator.setStrength(Collator.IDENTICAL);
    }
    
    /**
     * Construct a LocalizedComparatorFactory for the given locale and
     * the default strength {@link Collator#IDENTICAL}.
     * 
     * @param locale The local for which to contruct the string Collator.
     * 
     * @see Collator#getInstance(Locale)
     */
    public LocalizedComparatorFactory(Locale locale)
    {
        this.collator = Collator.getInstance(locale);
        this.collator.setStrength(Collator.IDENTICAL);
    }
    
    /**
     * Construct a LocalizedComparatorFactory for the given locale and
     * the default strength {@link Collator#IDENTICAL}.
     * 
     * @param locale The local for which to contruct the string Collator.
     * @param strength The strength of the underlying Collator. One of
     * <ul><li>{@link Collator#PRIMARY}</li>
     *     <li>{@link Collator#SECONDARY}</li>
     *     <li>{@link Collator#TERTIARY}</li>
     *     <li>{@link Collator#IDENTICAL}</li>
     * </ul>
     * 
     * @see Collator#getInstance(Locale)
     * @see Collator#setStrength(int)
     */
    public LocalizedComparatorFactory(Locale locale, int strength)
    {
        this.collator = Collator.getInstance(locale);
        this.collator.setStrength(strength);
    }

    /* (non-Javadoc)
     * @see org.clazzes.util.comparators.DefaultComparatorFactory#getComparator(java.lang.Class, boolean)
     */
    @Override
    public Comparator getComparator(Class clazz) {
       
        if (String.class.isAssignableFrom(clazz))
            return this.collator;
       
        return super.getComparator(clazz);
    }

    /**
     * @return The collator used for string comparisons.
     */
    public Collator getCollator() {
        return this.collator;
    }

    /**
     * @param collator The collator for string comparisons to set.
     */
    public void setCollator(Collator collator) {
        this.collator = collator;
    }
    
}
