/***********************************************************
 * $Id: POJOFieldComparator.java 87 2007-12-17 01:06:40Z lech $
 * 
 * Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: Feb 28, 2007
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.comparators;

import java.lang.reflect.InvocationTargetException;
import java.util.Comparator;

import org.clazzes.util.reflect.FieldAccessor;
import org.clazzes.util.reflect.POJOFieldAccessor;

/**
 * This comparator compares a POJO by the value of the supplied property.
 * 
 * @author wglas
 */
public class POJOFieldComparator implements Comparator {

    private FieldAccessor fieldAccessor;
    private Comparator fieldComparator;
    
    /**
     * Contructs a field comparator, which uses the given accessor and
     * Comparator.
     * 
     * @param fieldAccessor The field accessor, which is used to fetch field values.
     * @param fieldComparator The comparator, which compares field values.
     */
    public POJOFieldComparator(FieldAccessor fieldAccessor, Comparator fieldComparator)
    {
        this.fieldAccessor = fieldAccessor;
        this.fieldComparator = fieldComparator;
    }
    
    /**
     * Contructs a field comparator, which uses the given accessor and
     * constructs the comparator for the field using the given ComparatorFactory.
     * 
     * @param fieldAccessor The field accessor, which is used to fetch field values.
     * @param comparatorFactory The factory used to create a comparator base d on the type of the field.
     */
    public POJOFieldComparator(FieldAccessor fieldAccessor, ComparatorFactory comparatorFactory)
    {
        this.fieldAccessor = fieldAccessor;
        this.fieldComparator = comparatorFactory.getComparator(fieldAccessor.getType());
    }
    
    /**
     * Contructs a field comparator, which compare POJOs by the value
     * of the given property name using the given Comparator.
     * 
     * @param clazz The POJO clazz.
     * @param property The name of the POJO property, for which we use the getter function.
     * @param fieldComparator The underlying comparator.
     * @throws SecurityException If the security comtext does not allow acces to the getter.
     * @throws NoSuchMethodException If the geter method does not exist.
     * 
     * @see POJOFieldAccessor#POJOFieldAccessor(Class, String)
     */
    public POJOFieldComparator(Class clazz, String property, Comparator fieldComparator) throws SecurityException, NoSuchMethodException
    {
        this.fieldAccessor = new POJOFieldAccessor(clazz,property);
        this.fieldComparator = fieldComparator;        
    }
    
    /**
     * Contructs a field comparator, which comparea POJOs by the value
     * of the given property name and the comparator constructed using the given ComparatorFactory.
     * 
     * @param clazz The POJO clazz.
     * @param property The name of the POJO property, for which we use the getter function.
     * @param comparatorFactory The factory used to create a comparator base d on the type of the field.
     * @throws SecurityException If the security comtext does not allow acces to the getter.
     * @throws NoSuchMethodException If the geter method does not exist.
     * 
     * @see POJOFieldAccessor#POJOFieldAccessor(Class, String)
     */
    public POJOFieldComparator(Class clazz, String property, ComparatorFactory comparatorFactory) throws SecurityException, NoSuchMethodException
    {
        this.fieldAccessor = new POJOFieldAccessor(clazz,property);
        this.fieldComparator = comparatorFactory.getComparator(this.fieldAccessor.getType());   
    }
    
    /* (non-Javadoc)
     * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
     */
    public int compare(Object a, Object b) {
        
        if (a==null)
            return b==null ? 0 : -1;
        
        if (b==null)
            return 1;
        
        try {
            return this.fieldComparator.compare(this.fieldAccessor.getValueOf(a), this.fieldAccessor.getValueOf(b));
            
        } catch (IllegalArgumentException e) {
            throw new RuntimeException("Illegal argument fetching field ["+this.fieldAccessor.getName()+"] for comparison.",e);
        } catch (IllegalAccessException e) {
            throw new RuntimeException("Illegal access fetching field ["+this.fieldAccessor.getName()+"] for comparison.",e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException("Caught exception fetching field ["+this.fieldAccessor.getName()+"] for comparison.",e);
        }
    }

    /**
     * @return the fieldAccessor
     */
    public FieldAccessor getFieldAccessor() {
        return this.fieldAccessor;
    }

    /**
     * @param fieldAccessor the fieldAccessor to set
     */
    public void setFieldAccessor(FieldAccessor fieldAccessor) {
        this.fieldAccessor = fieldAccessor;
    }

    /**
     * @return the fieldComparator
     */
    public Comparator getFieldComparator() {
        return this.fieldComparator;
    }

    /**
     * @param fieldComparator the fieldComparator to set
     */
    public void setFieldComparator(Comparator fieldComparator) {
        this.fieldComparator = fieldComparator;
    }

}
