/***********************************************************
 * $Id: Util.java 87 2007-12-17 01:06:40Z lech $
 * 
 * Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 2006-03-13
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.lang;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.SecureRandom;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;
import java.util.zip.CRC32;

public class Util {

    /**
     * Enumeration of all OS we are able to detect
     */
    public enum OperatingSystem {
        UNKOWN, WINDOWS, UNIX
    };

    public final static String CHARSET_ASCII = "US-ASCII";

    public final static String CHARSET_ISO88591 = "ISO-8859-1";

    public final static String CHARSET_UTF8 = "UTF-8";

    /**
     * OS we are working in. Use
     * 
     * @getOperatingSystem() and
     * @inWindows() to avoid Unkown answers
     */
    private static OperatingSystem operating_system = OperatingSystem.UNKOWN;

    /**
     * Set to private, because this class defines no public non-static method.
     */
    private Util() {
    }

    /**
     * Compares a and b and returns true, if both values are null or
     * both are equal.
     * 
     * @param a The first object to compare.
     * @param b The second object to compare.
     * @return <code>true</code>, if <code>a==null && b==null</code>
     *         or <code>a.equals(b)</code>.
     */
    public static boolean equalsNullAware(Object a, Object b)
    {
        if (a == null)
            return b == null;
        return a.equals(b);
    }
    
    /**
     * Calculate a standard CRC32 Checksum using java.util.zip.CRC32.
     * 
     * @param data
     *            The Bytes to calculate the Checksum from.
     * @return The Checksum
     */
    public static int getCRC32(final byte[] data) {

        CRC32 crc = new CRC32();
        crc.update(data);

        return (int) crc.getValue();
    }

    /**
     * Fills up a Byte Array to a multiple of x Bytes. fillUp(byte[12], 16,
     * 0x00) will produce a byte[16], the last 4 Bytes are 0x00.
     * fillUp(byte[16], 16, 0x13) will produce a copy of the byte[16].
     * fillUp(byte[18], 16, 0x01) will produce a byte[32], the lase 14 Bytes are
     * 0x01. fillUp(byte[97], 16, 0xFF) will produce a byte[112], the last 15
     * Bytes are 0xFF.
     * 
     * @param src
     *            The Byte Array to fill up.
     * @param length
     *            The src ByteArray will be filled up to a length of n*length
     *            Bytes.
     * @param fillWith
     *            The value of the extra Bytes.
     * @return The filled up Byte Array.
     */
    public static byte[] fillUp(byte[] src, final int length,
            final byte fillWith) {
        int rest = src.length % length;
        if (rest == 0)
            return src;
        int newLength = src.length + length - rest;
        byte[] newSrc = new byte[newLength];
        System.arraycopy(src, 0, newSrc, 0, src.length);
        for (int i = src.length; i < newLength; i++)
            newSrc[i] = fillWith;
        return newSrc;
    }

    /**
     * Procudes an explicitely filled byte array with the given length. Useful
     * to produce "empty" arrays whose bytes are reliable set to 0.
     * 
     * @param length
     *            The length of the byte array to be created.
     * @param fillWith
     *            The value of all bytes of the new byte array.
     * @return The filled Byte Array.
     */
    public static byte[] filledByteArray(final int length, final byte fillWith) {
        byte[] ba = new byte[length];
        for (int i = 0; i < length; i++) {
            ba[i] = fillWith;
        }
        return ba;
    }

    public static Calendar getUTCnow() {
        return Calendar.getInstance(TimeZone.getTimeZone("UTC"));
    }

    public static Calendar getUTCtimestamp(long millis) {
        Calendar cal = GregorianCalendar.getInstance(TimeZone
                .getTimeZone("UTC"));
        cal.setTimeInMillis(millis);
        return cal;
    }

    /**
     * Get a Calendar-Object for the given date (starting at 00:00:00.000.
     * 
     * @param year
     *            The human readable year (2006)
     * @param month
     *            The human readable month (April = 4)
     * @param day
     *            The day of the month.
     * @return
     */
    public static Calendar getCalendar(final int year, final int month,
            final int day) {
        Calendar cal = GregorianCalendar.getInstance(TimeZone
                .getTimeZone("UTC"));
        cal.set(Calendar.YEAR, year);
        cal.set(Calendar.MONTH, month - 1);
        cal.set(Calendar.DAY_OF_MONTH, day);
        cal.set(Calendar.HOUR_OF_DAY, 0);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        return cal;
    }

    /**
     * Calculate the XOR between two Byte arrays;
     * 
     * @param a
     * @param b
     * @return
     * @throws Exception
     */
    public static byte[] xor(final byte[] a, final byte[] b) {
        if (a == null || b == null) {
            throw new NullPointerException("arrays must not be null.");
        } else {
            if (a.length != b.length) {
                throw new IllegalArgumentException(
                        "arrays must be of the same length. (" + a.length
                                + " != " + b.length + ")");
            }
        }
        int len = a.length;
        byte[] buffer = new byte[a.length];
        for (int i = 0; i < len; i++) {
            int v1 = a[i] & 0xff;
            int v2 = b[i] & 0xff;
            int xor = v1 ^ v2;
            buffer[i] = (byte) (xor & 0xff);
        }
        return buffer;
    }

    /**
     * Invert a Byte array.
     * 
     * @param a
     * @return
     * @throws Exception
     */
    public static byte[] invert(final byte[] a) {
        if (a == null) {
            throw new NullPointerException("Array must not be null.");
        }
        return (xor(a, filledByteArray(a.length, u8ToByte(0xFF))));
    }

    /**
     * Give me a SecureRandom of a desired length.
     * 
     * @param len
     *            The Length of the random value.
     * @return The random value.
     */
    public static byte[] getRandom(final int len) {
        byte[] random = new byte[len];
        try {
            SecureRandom srnd = SecureRandom.getInstance("SHA1PRNG");
            srnd.nextBytes(random);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return random;
    }

    /**
     * Give me the SHA1 Value of a Byte array.
     * 
     * @param input
     * @return
     * @throws Exception
     */
    public static byte[] getSHA1hash(final byte[] input) throws Exception {

        byte[] result = null;

        MessageDigest md = MessageDigest.getInstance("SHA1");
        md.update(input);
        result = md.digest();

        return result;
    }

    /**
     * Give me the SHA1 Value of a String.
     * 
     * @param input
     * @return
     * @throws Exception
     */
    public static String getSHA1hash(final String str) throws Exception {

        MessageDigest md = MessageDigest.getInstance("SHA1");
        md.update(str.getBytes());
        String result = Util.asHex(md.digest(), "");

        return result;
    }

    /**
     * Test if two Byte arrays are equal.
     * 
     * @param a
     * @param b
     * @return
     * @throws NullPointerException
     */
    public static boolean equals(final byte[] a, final byte[] b)
            throws NullPointerException {

        if (a == null || b == null) {
            throw new NullPointerException("arrays must not be null.");
        } else {
            if (a.length == b.length) {

                for (int i = 0; i < a.length; i++) {
                    if (a[i] != b[i]) {
                        return false;
                    }
                }
                return true;

            } else {
                return false;
            }
        }
    }

    /**
     * Copy two Byte arrays into a single resulting Byte array.
     * 
     * @param a
     * @param b
     * @return
     */
    public static byte[] concat(final byte[] a, final byte[] b) {

        byte[] result = new byte[a.length + b.length];
        System.arraycopy(a, 0, result, 0, a.length);
        System.arraycopy(b, 0, result, a.length, b.length);

        return result;
    }

    public static byte[] subArray(final byte[] src, final int offset,
            final int length) {

        byte[] result = new byte[length];
        System.arraycopy(src, offset, result, 0, length);

        return result;
    }

    /**
     * Turns array of bytes into string.
     * 
     * @param buf
     *            Array of bytes to convert to hex string, with a space between
     *            each byte.
     * @return Generated hex string (without any 0x prefix).
     */
    public static String asHex(final byte buf[]) {
        return asHex(buf, " ");
    }

    /**
     * Turns array of bytes into string.
     * 
     * @param buf
     *            Array of bytes to convert to hex string.
     * @param sep
     *            Separator to put between each byte, null stands for ""
     * @return Generated hex string (without any 0x prefix).
     */
    public static String asHex(final byte buf[], String sep) {
        if (buf == null)
            return "null";

        StringBuffer strbuf = new StringBuffer(buf.length * 2);
        int i;

        for (i = 0; i < buf.length; i++) {
            if (((int) buf[i] & 0xff) < 0x10)
                strbuf.append("0");

            strbuf.append(Long.toString((int) buf[i] & 0xff, 16));
            if (i < (buf.length - 1) && sep != null)
                strbuf.append(sep);
        }

        //log.debug(strbuf.toString());

        return strbuf.toString();
    }

    /**
     * Decodes a string contain hex digits to a byte[]
     * 
     * @param s
     *            The string to decode.
     * @return The decoded byte[]
     * @throws some
     *             runtimeexceptions if the string is invalid
     */
    public static byte[] decodeHexstring(String s) {
        String s2 = s.replaceAll(" ", "").replaceFirst("0x", ""); // we want
                                                                    // accept
                                                                    // "0x 01 02
                                                                    // 03"
        int bal = s2.length() / 2;
        byte[] ba = new byte[bal];
        for (int i = 0; i < bal; i++) {
            String sub = s2.substring(i * 2, i * 2 + 2);
            ba[i] = u8ToByte(Short.parseShort(sub, 16));
        }
        return ba;
    }

    /**
     * Converts a (pseudo-signed) byte to an unsigned 8 bit value.
     * 
     * @param u
     * @return
     */
    public static int byteToU8(byte b) {
        int result = 0;
        result = b & 0x7F;
        if (b < 0)
            result += 0x80;
        return result;
    }

    /**
     * Converts an unsigned 8 bit value to a (pseudo-signed) byte.
     * 
     * @param u
     * @return
     */
    public static byte u8ToByte(int u) {
        // if (logger.isDebugEnabled()) logger.debug(String.format("u8ToByte(%d=0x%04x) called", u, u));
        int high = u & 0x80;
        int low = u & 0x7F;
        byte b = (new Integer(low)).byteValue();
        if (high > 0)
            b -= 128;
        return b;
    }

    /**
     * Converts a String object to an byte[] containing the ASCII coded string
     * and a 0 byte.
     * 
     * @param s
     * @return
     */
    public static byte[] createZeroTerminatedString(String s) {
        return createZeroTerminatedString(s, CHARSET_ASCII);
    }

    /**
     * Converts a String object to an byte[] containing the string coded with
     * the given charset and a 0 byte.
     * 
     * @param s
     * @param charset
     *            We recommend to use this class' CHARSET_* constants
     * @return
     */
    public static byte[] createZeroTerminatedString(String s, String charset) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream((s == null ? 0
                : s.length() + 1));
        try {
            if (s != null)
                baos.write(s.getBytes(charset));
        } catch (UnsupportedEncodingException e) {
        } catch (IOException e) {
        }
        baos.write((byte) 0);
        return baos.toByteArray();
    }

    /**
     * Converts a zero terminated ASCII string to a java String object,
     * suggesting ASCII charset.
     * 
     * @param ba
     * @return
     */
    public static String parseZeroTerminatedString(byte[] ba) {
        return parseZeroTerminatedString(ba, CHARSET_ASCII);
    }

    /**
     * Converts a zero terminated ASCII string to a java String object, using
     * the given charset.
     * 
     * @param ba
     * @param charset
     *            We recommend to use this class' CHARSET_* constants
     * @return
     */
    public static String parseZeroTerminatedString(byte[] ba, String charset) {
        String s = null;
        try {
            if (ba != null && ba.length > 1) {
                // find 0
                int zi = 0;
                while (zi < ba.length) {
                    if (ba[zi] == 0)
                        break;
                    zi++;
                }
                // if 0 was not found, zi=ba.length+1, this is ok
                s = new String(ba, 0, zi, charset);
            }
        } catch (UnsupportedEncodingException e) {
        }
        return s;
    }

    /**
     * Returns which type of OS we have, after dedecting if necessary
     * 
     * @return
     */
    public static OperatingSystem getOperatingSystem() {
        if (operating_system == OperatingSystem.UNKOWN) {
            String osName = System.getProperty("os.name");
            if (osName.toLowerCase().contains("window")) // !windows
            {
                operating_system = OperatingSystem.WINDOWS;
            } else {
                operating_system = OperatingSystem.UNIX;
            }

        }
        return operating_system;
    }

    /**
     * Tells if we are under M$ Windows, after checking if necessary
     * 
     * @return
     */
    public static boolean inWindows() {
        return (getOperatingSystem() == OperatingSystem.UNKOWN);
    }

    /**
     * Computes an byte[2] containing the BCD representation of the given int
     * 
     * @param pin
     *            The pin to encodde, 1 .. 9999
     * @return
     */
    public static byte[] asBCD16(int pin) {
        byte[] ba = new byte[2];
        int bcd = 0;
        bcd |= (pin % 10);
        pin /= 10;
        bcd |= (pin % 10) << 4;
        pin /= 10;
        bcd |= (pin % 10) << 8;
        pin /= 10;
        bcd |= (pin % 10) << 12;
        ba[0] = (byte) ((bcd & 0xFF00) >> 8);
        ba[1] = (byte) ((bcd & 0x00FF));
        return ba;
    }

    /**
     * Decodes a BCD coded integer
     * 
     * @param ba
     *            The bytearray containing the BCD bytes
     * @return The decoded value
     */
    public static int decodeBCD(byte[] ba) {
        int ret = 0;
        for (int i = 0; i < ba.length; i++) {
            ret *= 100;
            ret += ((ba[i] & 0xF0) >> 4) * 10 + (ba[i] & 0x0F);
        }
        return ret;
    }


}
